import java.awt.*;
import java.util.*;
import java.io.*;

// class CheckerProgram
// does the move I/O and is what you should run to use our agent
public class CheckerProgram {
    final static int UNDO = -1;
    final static int EXIT = -2;
    final static int TAKEBACK = -3;

    // checkers object, holds the board
    static Checkers c = new Checkers();

    // flag for undoing a move
    static boolean undoMove = false;

    // number of moves made
    static int numMoves = 0;

    public static void main (String[] args) throws IOException {
	
	// hold the last two boards for undo
	int[][] lastBoard = Checkers.GetBoardCopy (c.board);
	int[][] undoBoard = Checkers.GetBoardCopy (c.board);

	int turn;
	int[] move;
	
	// set the initial board state
	c.Reset();

	// find out what color we are 
	BufferedReader br = new BufferedReader (new InputStreamReader (System.in));
	System.out.println ("One Jump Behind");
	System.out.println ("---------------");
	System.out.print ("Am I red or white (r/w)? ");
	String color = br.readLine();
	System.out.println();
	
	// if we are white, let the human go first

	if (color.equals ("w")) turn = Checkers.WHITE;
	else turn = Checkers.RED;

	System.out.print ("Am I going first (y/n)? ");
	String first = br.readLine();
	System.out.println();

	if (!first.equals("y") && !first.equals("Y"))
	{
		turn = Checkers.GetOpponent(turn);
		lastBoard = Checkers.GetBoardCopy (c.board);
		move = HumanInput (lastBoard);
		c.PrintMove (move, turn);
		c.UpdateBoard (c.board, move, turn);
		turn = Checkers.GetOpponent(turn);
	}

	// enter a loop where the turns switch off given the colors above
	while (c.AIMove (c.board, turn))
	    {
		numMoves++;
		turn = Checkers.GetOpponent(turn);

		undoBoard = Checkers.GetBoardCopy (lastBoard);
		lastBoard = Checkers.GetBoardCopy (c.board);

		// check if the human won
		if ((Checkers.GetMoves(c.board, turn)).size() == 0)
		    {
			System.out.println ("You win...");
			System.exit(0);
		    }
		// get the human's input
		move = HumanInput (undoBoard);
		
		// print out the human's move and update the board accordingly
		c.PrintMove (move, turn);
		c.UpdateBoard (c.board, move, turn);
		
		turn = Checkers.GetOpponent(turn);
	    }		

	System.out.println ("Number of moves: " + numMoves);
	
    }

    // get human input
    public static int[] HumanInput(int[][] undoBoard) throws IOException
    {
	int[] move;

	int[][] takeback = Checkers.GetBoardCopy(c.board);

	// this loop is for undo
	do
	    {
		c.Print (c.board);
		undoMove = false;
		
		// get the actual input from the keyboard
		move = GetInputMove();

		if (move[0] == EXIT)
		    {
			System.out.println ("Exit");
			System.exit(0);
		    }
		else if (move[0] == UNDO)
		    {
			undoMove = true;
			c.board = Checkers.GetBoardCopy (undoBoard);
			numMoves--;
		    }
		else if (move[0] == TAKEBACK)
		   {
			undoMove = true;
		   }
	    } while (undoMove);
	return move;
    }
    
    // input a move
    public static int[] GetInputMove () throws IOException
    {
	int[] move = new int[4];
	move[0] = EXIT;
	boolean noException = true;
	
	InputStreamReader ir = new InputStreamReader (System.in);
	BufferedReader br = new BufferedReader (ir);
	
	int sp = 0, ep = 0;
	String startPos, endPos;

	// get the start position, error checking for any non-integers
	// "exit" is okay, it exits the program
	do
	    {
		noException = true;

		System.out.print ("Enter start position: ");
		startPos =  br.readLine();
		
		// check for exit or undo
		if (startPos.equals("exit")) return move;
		else if (startPos.equals("undo"))
		    {
			move[0] = UNDO;
			return move;
		    }
	        else if (startPos.equals("takeback"))
		    { 
			move[0] = TAKEBACK;
			return move;
		    }
		try
		    {
			sp = Integer.parseInt (startPos);
		    }
	    	catch (NumberFormatException e) { System.out.println ("Incorrect input, try again..."); noException = false; }
	    } while (!noException);
	
	// now get the next position and check for non-integer inputs
	do
	    {
		noException = true;
		System.out.print ("Enter next position: ");
		endPos = br.readLine();

		try {
		    ep = Integer.parseInt (endPos);
		}
		catch (NumberFormatException e) { System.out.println ("Incorrect input, try again..."); noException = false; }
	    } while (!noException);
	
	// given the numbers put in, calculate the move according to array indices
	move[0] = (int)((sp-1)/4);
	move[1] = sp - 1 - move[0]*4;
	move[2] = (int)((ep-1)/4);
	move[3] = ep - 1 - move[2]*4;

	boolean done = false;
	int newMove[];

	// now get the rest of the inputs
	// this outer loop keeps going until "done" is inputted, so multiple jumps can be entered
	do
	    {
		// this loop is for error checking in the input
		do
		    {
			noException = true;

			// get the input
			System.out.print ("Enter next position (or done): ");
			endPos = br.readLine();

			// no more input, so stop
			if (endPos.equals("done")) 
			    {
				done = true;
			    }
			// undo, return it so we can roll back
			else if (endPos.equals("undo"))
			    {
				move[0] = UNDO;
				return move;
			    }
			else if (endPos.equals("takeback"))
			    { 
				move[0] = TAKEBACK;
				return move;
			    }
			else
			    {
				// check for a non-integer input
				try
				    {
					ep = Integer.parseInt (endPos);
					newMove = new int[move.length + 2];
					System.arraycopy (move, 0, newMove, 0, move.length);
					newMove[move.length] = (int)((ep-1)/4);
					newMove[move.length+1] = ep - 1 - newMove[move.length]*4;
					move = newMove;
				    }
				catch (NumberFormatException e) { System.out.println ("Incorrect input, try again..."); noException = false; }  
			    }
		    } while (!noException);
	    } while (!done);
	
	return move;
    }
}
