import java.awt.*;
import java.util.*;

class Search {
    
    static final int MAXDEPTH = 10;
    private static int depthChange=0;
    //static final long MAXTIME = 5000;
    static final int EMPTY= Checkers.EMPTY;
    static final int WHITE = Checkers.WHITE;
    static final int RED= Checkers.RED;
    static final int WKING= Checkers.WHITEKING;
    static final int RKING= Checkers.REDKING;
    //value of kings and individual checkers
    final static int SINGLE = 150;
    final static int KING =   250;
    //importance of position (edges are better)
    final static int POSITION= 15;
    //worth of balance
    static final int BALANCE=50;
    
    //penalty for kings stuck on edges or corners
    static final int KCORN=25; 
    //position values
    static final int[][] REDPOSVAL= {{-3, -3, -3, -3},
                                     {-2, -2, -2, -2},
                     {-2, -1, -1, -2},
                     { 0, -2, -2,  0},
                     { 0, -2, -2,  0},
                     { 0,  0,  0,  0},
                     { 3,  2,  2,  3},
                     { 3,  0,  0,  3}};

    static final int[][] WHITEPOSVAL= {{3,0,0,3},
                       {3,2,2,3},
                       {0,0,0,0},
                       {0,-2,-2,0},
                       {0,-2,-2,0},
                       {-2,-1,-1,-2},
                       {-2,-2,-2,-2},
                       {-3,-3,-3,-3}};
    
    //value of shadow heuristic
    final static int SHADOW=5;

    public static int Max (int a, int b)
    {
    return (a > b ? a : b);
    }

    public static int Min (int a, int b)
    {
    return (a < b ? a : b);
    }

    // evaluate a board
    // red wants a high score, white wants a low score
    //awards each checker and king a value and implements
    //other heuristics
    public static int Evaluate (int[][] board)
    {
    int score = 0;
    int numRed = 0, numWhite = 0;
    int immobleRed=0, immobleWhite=0;
    for (int r=0; r<8; r++)
        {
        for (int c=0; c<4; c++)
        {
            if (board[r][c] == Checkers.RED) 
            {
                score -= SINGLE;
                score += positionValue(r, c, RED);
                if(immoble(board, r, c, RED))// immoble checkers twice as bad
                    immobleRed++;
               
        numRed++;
            }
            else if (board[r][c] == Checkers.REDKING)
            {
                score -= KING;
                score += positionValue(r, c, RKING);
                if(immoble(board,r, c, RED))
                    immobleRed++;
                
        numRed++;
            }
            else if (board[r][c] == Checkers.WHITE)
            {
                score += SINGLE;
                score -=positionValue(r, c, WHITE);
                if(immoble(board, r, c, WHITE))
                    immobleWhite++;
                
        numWhite++;
            }
            else if (board[r][c] == Checkers.WHITEKING)
            {
                score += KING;
                score -=positionValue(r, c, WKING);
                if(immoble(board, r, c, WHITE))
                    immobleWhite++;
                 
        numWhite++;
            }
        }
        }
       
       //if one side has no pieces, it is a goal state,
       // return highest possible value
    if (numRed == 0) return Integer.MAX_VALUE;
    else if (numWhite == 0) return -Integer.MAX_VALUE;
    
    //take away value for immobile pieces:
    //only do if there are lots of pieces
    if (numRed+numWhite >15)
    {
        score+= (SINGLE*immobleWhite);
        score-= (SINGLE*immobleRed);
    }

    //if (numRed < 6) score -= 50 * numRed;
    //if (numWhite < 6) score += 50 * numWhite;

    //empty spaces behind our lines
    score += shadow(RED, board);
    score += shadow(WHITE, board);
    //balance-- on balance is better
    score +=balance(board);
    
    return score;
    }
    
     public static boolean immoble (int[][] b, int r, int c, int turn)
    {
    //Vector moves = new Vector();
    //int[] move;
    
    if (r != 7 && (turn == RED || b[r][c] == WKING))
        {
        if (b[r+1][c] == EMPTY)
            {
            return false;
            }
        
        // in an odd row, can move down one row and
        // same column or one column left, unless all the way left already
        if (r % 2 == 1)
            {
            if (c != 0 && b[r+1][c-1] == EMPTY)
                {
                return false;
                }
            }
            
        // in an even row, can move down one row and same column
        // or one column right, unless all the way right already
        else if (c != 3 && b[r+1][c+1] == EMPTY) 
            {
            return false;
            }   
        }
    if (r != 0 && (turn == WHITE || b[r][c] == RKING))
        {
        // can alway move one row up and keep same column
        // (if it's empty that is)
        if (b[r-1][c] == EMPTY)
            {
            return false;
            }

        // in odd row, can move up one row and same column or
        // one column left, unless all the way left already
        if (r % 2 == 1)
            {
            if (c != 0 && b[r-1][c-1] == EMPTY)
                {
                return false;
                }
            }
        // in an even row, can move up one row and same column or
        // one column right, unless all the way right already (c=3)
        else if (c != 3 && b[r-1][c+1] == EMPTY)
            {
            return false;
            }
           
        }
    return true;           
    
    }
    
    
    
    private static int positionValue(int row, int column, int color)
    {
        if (color== RED)
            return REDPOSVAL[row][column]*POSITION;// if red, return the position value 
                                                // multiplied by the weight currently 
                                                // assigned to the position heuristic
                                                //otherwise return the negation of that value
                                                //for white
                                               
       else
       if (color== WHITE)
            return WHITEPOSVAL[row][column]*POSITION;
       int kingval=0;
       if (color == RKING)
       {
            if (column==0 || column == 3)
                kingval+= -KCORN;
            if (row == 0 || row == 7)
                kingval+= -KCORN;
       }
       if (color == WKING)
       {
            if (column==0 || column == 3)
                kingval+= KCORN;
            if (row == 0 || row == 7)
                kingval+= KCORN;
       }
       return kingval;
    }
    
    //heuristic for "shadow"-- empty spaces behind our pieces.
    // this is a good thing as it means our pieces can be jumped more easily
    private static int shadow(int color, int[][] board)
    {
        int score=0;
        int end;
        int inc;
        int jstart;
        int maxRed=0;
        int maxWhite=7;
        int numRedShadow=0;
        int numWhiteShadow=0;
        int minWhite=0;
        int minRed=0;
        if (color==RED || color==RKING)
        {
            inc=1;
            jstart=0;
            end=4;
            
        }
        else
        {
            jstart=4;
            end=8;
            inc = 1;
        }
            for (int i=0;i<4;i++)
                for (int j=jstart;j<end;j+=inc)
                    if (end==3 && board[j][i]== EMPTY)
                        numRedShadow++;
                    else 
                    if (end==4 && (board[j][i] == RED || board [j][i] == RKING))
                    {
                        minRed= Min(minRed, j);
                        maxRed= Max(maxRed, j);
                    }
                    else 
                    if (end==8 && board[j][i]== EMPTY)
                        numWhiteShadow++;
                    else 
                    if (end==8 && (board[j][i] == WHITE || board [j][i] == WKING))
                    {
                        maxWhite= Min(maxWhite, j);
                        minWhite= Max(minWhite, j);
                    }
        //ajust so shadows are only counted if there are pieces infront of it    
        
        numRedShadow= Min(numRedShadow, maxRed*4);
        numWhiteShadow= Min(numWhiteShadow, (8-maxWhite)*4);
        
        score = (numRedShadow-numWhiteShadow)*SHADOW;
        return score;
    }
    private static int balance(int[][] board)
    {
        //8by4
        int redbalance=0;
        int whitebalance=0;
        for (int i=0;i<8;i++)
            for (int j=0;j<4;j++)
            {
                if(j<2 && board[i][j]== WHITE || board [i][j]==WKING)
                    whitebalance--;
                else 
                if(j<1 && board[i][j]== WHITE || board [i][j]==WKING)
                    whitebalance++;
                 else   
                 if(j<2 && board[i][j]== RED || board [i][j]==RKING)
                    redbalance--;
                else 
                if(j<1 && board[i][j]== RED || board [i][j]==RKING)
                    redbalance++;
            }
            //returns positive values if white is less balanced
            //returns negative values if red is less balanced
            return (Math.abs(whitebalance)-Math.abs(redbalance))*BALANCE;  
        }

    // maxValue, part of minimax
    public static int MaxValue (int[][] board, int depth, int[] bestIndex, int turn, int alpha, int beta)
    {
    int bestScore = -Integer.MAX_VALUE;
    int newScore = 0;
    Vector moves = new Vector();
    
    /*
    if (depth >= MAXDEPTH)
        {
        maxReached = true;
        long t = (new Date()).getTime();
        if (t - time > MAXTIME) return Evaluate (board);
        }
    */
    // check for a cutoff node, and returns its heuristic value
    if (depth == MAXDEPTH) return Evaluate (board);

    // get all possible moves for this board
    //System.out.println("Max's moves:");
    moves = Checkers.GetMoves (board, turn);

    // if max has no moves then this state is a win for red, so return a huge number
    if (moves.size() == 0) return Integer.MAX_VALUE;

    int opponent = Checkers.GetOpponent(turn);
    
    // loop through all the moves and call minvalue on them
    for (int i=0; i < moves.size(); i++)
        {
       

        // make a temporary board to reflect the new move
        int tempBoard[][] = new int[8][4];
        tempBoard = Checkers.GetBoardCopy (board);
        Checkers.UpdateBoard (tempBoard, (int[])moves.get(i), turn);
        int[] temp = new int[1];
        newScore = Max (bestScore, MinValue (tempBoard, depth+1, temp, opponent, alpha, beta));
        
        if (newScore > bestScore)
            {
            bestScore = newScore;
            bestIndex[0] = i;

            // alpha-beta cutoff
            if (bestScore >= beta)
                {
                return bestScore;           
            }
                alpha = Max (alpha, bestScore);
        }
        }
    
    return bestScore;
    }   

    // minvalue, the other part of minimax
    public static int MinValue (int[][] board, int depth, int[] bestIndex, int turn, int alpha, int beta)
    {
    int bestScore = Integer.MAX_VALUE;
    int newScore;
    Vector moves = new Vector();
    
    /*
    if (depth >= MAXDEPTH)
        {
        maxReached = true;
        long t = (new Date()).getTime();
        if (t - time > MAXTIME) return Evaluate (board);
        }
    */

    // cutoff test, return the heuristic of a board if at max depth
    if (depth == MAXDEPTH) return Evaluate (board);

    // get all possible moves for this board
    moves = Checkers.GetMoves (board, turn);

    // if min has no moves it is a win for white, so return a very large negative number
    if (moves.size() == 0) return -Integer.MAX_VALUE;

    int opponent = Checkers.GetOpponent(turn);

    // loop through all the moves and call minvalue on them
    for (int i=0; i < moves.size(); i++)
        {
        /*
        if (maxReached == true && (new Date()).getTime() - time > MAXTIME)
            {
            return Integer.MAX_VALUE;
            }
        */

        // make and update a temporary board to reflect each move
        int tempBoard[][] = new int[8][4];
        tempBoard = Checkers.GetBoardCopy (board);
        Checkers.UpdateBoard (tempBoard, (int[])moves.get(i), turn);
        int[] temp = new int[1];
        newScore = Min (bestScore, MaxValue (tempBoard, depth+1, temp, opponent, alpha, beta));

        if (newScore < bestScore)
            {
            bestScore = newScore;
            bestIndex[0] = i;

            // alpha-beta cutoff
            if (bestScore <= alpha)
                { 
                return bestScore;
             }
            beta = Min (beta, bestScore);
            }
        }        
    
    return bestScore;
    }   

}

